@@ -1,4 +1,4 @@
-from pyrsistent import pmap
+from rpds import HashTrieMap
 import pytest
 
 from referencing import Anchor, Registry, Resource, Specification, exceptions
@@ -56,6 +56,30 @@ class TestRegistry:
             resource=two,
         )
 
+    def test_matmul_resource(self):
+        uri = "urn:example:resource"
+        resource = ID_AND_CHILDREN.create_resource({"ID": uri, "foo": 12})
+        registry = resource @ Registry()
+        assert registry == Registry().with_resource(uri, resource)
+
+    def test_matmul_many_resources(self):
+        one_uri = "urn:example:one"
+        one = ID_AND_CHILDREN.create_resource({"ID": one_uri, "foo": 12})
+
+        two_uri = "urn:example:two"
+        two = ID_AND_CHILDREN.create_resource({"ID": two_uri, "foo": 12})
+
+        registry = [one, two] @ Registry()
+        assert registry == Registry().with_resources(
+            [(one_uri, one), (two_uri, two)],
+        )
+
+    def test_matmul_resource_without_id(self):
+        resource = Resource.opaque(contents={"foo": "bar"})
+        with pytest.raises(exceptions.NoInternalID) as e:
+            resource @ Registry()
+        assert e.value == exceptions.NoInternalID(resource=resource)
+
     def test_with_contents_from_json_schema(self):
         uri = "urn:example"
         schema = {"$schema": "https://json-schema.org/draft/2020-12/schema"}
@@ -98,7 +122,7 @@ class TestRegistry:
         root = ID_AND_CHILDREN.create_resource(
             {"ID": "urn:root", "children": [{"ID": child_id, "foo": 12}]},
         )
-        registry = Registry().with_resource(root.id(), root)
+        registry = root @ Registry()
         with pytest.raises(LookupError):
             registry[child_id]
 
@@ -109,11 +133,9 @@ class TestRegistry:
         resource = ID_AND_CHILDREN.create_resource(
             {"ID": "urn:bar", "anchors": {"foo": 12}},
         )
-        registry = Registry().with_resource(resource.id(), resource)
-        with pytest.raises(LookupError):
-            registry.anchor(resource.id(), "foo")
+        registry = resource @ Registry()
 
-        assert registry.crawl().anchor(resource.id(), "foo") == Anchor(
+        assert registry.crawl().anchor(resource.id(), "foo").value == Anchor(
             name="foo",
             resource=ID_AND_CHILDREN.create_resource(12),
         )
@@ -121,10 +143,8 @@ class TestRegistry:
     def test_crawl_finds_anchors_no_id(self):
         resource = ID_AND_CHILDREN.create_resource({"anchors": {"foo": 12}})
         registry = Registry().with_resource("urn:root", resource)
-        with pytest.raises(LookupError):
-            registry.anchor("urn:root", "foo")
 
-        assert registry.crawl().anchor("urn:root", "foo") == Anchor(
+        assert registry.crawl().anchor("urn:root", "foo").value == Anchor(
             name="foo",
             resource=ID_AND_CHILDREN.create_resource(12),
         )
@@ -135,6 +155,16 @@ class TestRegistry:
         registry = Registry().with_resource(uri, resource)
         assert registry.contents(uri) == {"foo": "bar"}
 
+    def test_crawled_anchor(self):
+        resource = ID_AND_CHILDREN.create_resource({"anchors": {"foo": "bar"}})
+        registry = Registry().with_resource("urn:example", resource)
+        retrieved = registry.anchor("urn:example", "foo")
+        assert retrieved.value == Anchor(
+            name="foo",
+            resource=ID_AND_CHILDREN.create_resource("bar"),
+        )
+        assert retrieved.registry == registry.crawl()
+
     def test_init(self):
         one = Resource.opaque(contents={})
         two = ID_AND_CHILDREN.create_resource({"foo": "bar"})
@@ -158,7 +188,7 @@ class TestRegistry:
 
     def test_dict_conversion(self):
         """
-        Passing a `dict` to `Registry` gets converted to a `pmap`.
+        Passing a `dict` to `Registry` gets converted to a `HashTrieMap`.
 
         So continuing to use the registry works.
         """
@@ -167,7 +197,7 @@ class TestRegistry:
         two = ID_AND_CHILDREN.create_resource({"foo": "bar"})
         registry = Registry(
             {"http://example.com/1": one},
-        ).with_resources([("http://example.com/foo/bar", two)])
+        ).with_resource("http://example.com/foo/bar", two)
         assert (
             registry.crawl()
             == Registry()
@@ -214,7 +244,7 @@ class TestRegistry:
                 ("http://example.com/baz", three),
                 ("http://example.com/foo/quux", four),
             ],
-            anchors=pmap(
+            anchors=HashTrieMap(
                 {
                     ("http://example.com/foo/quux", "foo"): Anchor(
                         name="foo",
@@ -401,29 +431,33 @@ class TestRegistry:
     def test_retrieve(self):
         foo = Resource.opaque({"foo": "bar"})
         registry = Registry(retrieve=lambda uri: foo)
-        assert registry["urn:example"] == foo
+        assert registry.get_or_retrieve("urn:example").value == foo
 
     def test_retrieve_arbitrary_exception(self):
+        foo = Resource.opaque({"foo": "bar"})
+
         def retrieve(uri):
             if uri == "urn:succeed":
-                return {}
+                return foo
             raise Exception("Oh no!")
 
         registry = Registry(retrieve=retrieve)
-        assert registry["urn:succeed"] == {}
+        assert registry.get_or_retrieve("urn:succeed").value == foo
         with pytest.raises(exceptions.Unretrievable):
-            registry["urn:uhoh"]
+            registry.get_or_retrieve("urn:uhoh")
 
     def test_retrieve_no_such_resource(self):
+        foo = Resource.opaque({"foo": "bar"})
+
         def retrieve(uri):
             if uri == "urn:succeed":
-                return {}
+                return foo
             raise exceptions.NoSuchResource(ref=uri)
 
         registry = Registry(retrieve=retrieve)
-        assert registry["urn:succeed"] == {}
+        assert registry.get_or_retrieve("urn:succeed").value == foo
         with pytest.raises(exceptions.NoSuchResource):
-            registry["urn:uhoh"]
+            registry.get_or_retrieve("urn:uhoh")
 
     def test_retrieve_cannot_determine_specification(self):
         def retrieve(uri):
@@ -431,7 +465,7 @@ class TestRegistry:
 
         registry = Registry(retrieve=retrieve)
         with pytest.raises(exceptions.CannotDetermineSpecification):
-            registry["urn:uhoh"]
+            registry.get_or_retrieve("urn:uhoh")
 
     def test_retrieve_already_available_resource(self):
         def retrieve(uri):
@@ -440,6 +474,7 @@ class TestRegistry:
         foo = Resource.opaque({"foo": "bar"})
         registry = Registry({"urn:example": foo})
         assert registry["urn:example"] == foo
+        assert registry.get_or_retrieve("urn:example").value == foo
 
     def test_retrieve_first_checks_crawlable_resource(self):
         def retrieve(uri):
@@ -626,8 +661,8 @@ class TestResolver:
                 ],
             },
         )
-        resolver = Registry().with_resource(root.id(), root).resolver()
-        resolved = resolver.lookup("http://example.com/a")
+        registry = root @ Registry()
+        resolved = registry.resolver().lookup("http://example.com/a")
         assert resolved.contents == {"ID": "http://example.com/a", "foo": 12}
 
     def test_lookup_anchor_with_id(self):
@@ -637,8 +672,8 @@ class TestResolver:
                 "anchors": {"foo": 12},
             },
         )
-        resolver = Registry().with_resource(root.id(), root).resolver()
-        resolved = resolver.lookup("http://example.com/#foo")
+        registry = root @ Registry()
+        resolved = registry.resolver().lookup("http://example.com/#foo")
         assert resolved.contents == 12
 
     def test_lookup_anchor_without_id(self):
@@ -687,7 +722,7 @@ class TestResolver:
         with pytest.raises(exceptions.Unresolvable) as e:
             resolver.lookup(ref)
         assert e.value == exceptions.NoSuchAnchor(
-            ref=ref,
+            ref="urn:example",
             resource=root,
             anchor="noSuchAnchor",
         )
@@ -698,6 +733,38 @@ class TestResolver:
         resolved = resolver.lookup("http://example.com/")
         assert resolved.contents == resource.contents
 
+    def test_repeated_lookup_from_retrieved_resource(self):
+        """
+        A (custom-)retrieved resource is added to the registry returned by
+        looking it up.
+        """
+        resource = Resource.opaque(contents={"foo": "baz"})
+        once = [resource]
+
+        def retrieve(uri: str):
+            return once.pop()
+
+        resolver = Registry(retrieve=retrieve).resolver()
+        resolved = resolver.lookup("http://example.com/")
+        assert resolved.contents == resource.contents
+
+        resolved = resolved.resolver.lookup("http://example.com/")
+        assert resolved.contents == resource.contents
+
+    def test_repeated_anchor_lookup_from_retrieved_resource(self):
+        resource = Resource.opaque(contents={"foo": "baz"})
+        once = [resource]
+
+        def retrieve(uri: str):
+            return once.pop()
+
+        resolver = Registry(retrieve=retrieve).resolver()
+        resolved = resolver.lookup("http://example.com/")
+        assert resolved.contents == resource.contents
+
+        resolved = resolved.resolver.lookup("#")
+        assert resolved.contents == resource.contents
+
     # FIXME: The tests below aren't really representable in the current
     #        suite, though we should probably think of ways to do so.
 
@@ -713,7 +780,7 @@ class TestResolver:
                 ],
             },
         )
-        registry = Registry().with_resource(root.id(), root)
+        registry = root @ Registry()
 
         resolver = registry.resolver()
         first = resolver.lookup("http://example.com/")
@@ -740,7 +807,7 @@ class TestResolver:
                 ],
             },
         )
-        registry = Registry().with_resource(root.id(), root)
+        registry = root @ Registry()
 
         resolver = registry.resolver()
         first = resolver.lookup("http://example.com/")
@@ -771,9 +838,7 @@ class TestResolver:
                 "children": [{"ID": "two-child/"}],
             },
         )
-        registry = Registry().with_resources(
-            [(one.id(), one), (two.id(), two)],
-        )
+        registry = [one, two] @ Registry()
 
         resolver = registry.resolver()
         first = resolver.lookup("http://example.com/")
